import datetime
import shutil
from pathlib import Path
from textwrap import dedent

import pytest
import pytz
import toml
from git import Repo

from metric_config_parser import AnalysisUnit
from metric_config_parser.analysis import AnalysisSpec
from metric_config_parser.config import (
    Config,
    ConfigCollection,
    DefaultConfig,
    DefinitionConfig,
    LocalConfigCollection,
    Outcome,
)
from metric_config_parser.data_source import DataSourceJoinRelationship
from metric_config_parser.errors import DefinitionNotFound
from metric_config_parser.metric import MetricLevel
from metric_config_parser.outcome import OutcomeSpec

TEST_DIR = Path(__file__).parent


class TestConfigIntegration:
    config_str = dedent(
        """
        [metrics]
        weekly = ["active_hours"]

        [metrics.active_hours.statistics.bootstrap_mean]
        """
    )
    spec = AnalysisSpec.from_dict(toml.loads(config_str))

    def test_old_config(self):
        config = Config(
            slug="new_table",
            spec=self.spec,
            last_modified=pytz.UTC.localize(
                datetime.datetime.utcnow() - datetime.timedelta(days=1)
            ),
        )

        config_collection = ConfigCollection([config])

        assert config_collection.spec_for_experiment("new_table") is not None
        assert config_collection.spec_for_outcome("test", "foo") is None
        assert config_collection.get_platform_defaults("desktop") is None
        assert config_collection.get_segment_data_source_definition("foo", "test") is None

    def test_definition_config(self):
        config_str = dedent(
            """
            [metrics.retained]
            select_expression = "COALESCE(COUNT(document_id), 0) > 0"
            data_source = "baseline"

            [data_sources.baseline]
            from_expression = "mozdata.search.baseline"
            experiments_column_type = "simple"
            """
        )
        definition = DefinitionConfig(
            slug="firefox_desktop",
            platform="firefox_desktop",
            spec=AnalysisSpec.from_dict(toml.loads(config_str)),
            last_modified=datetime.datetime.now(),
        )

        assert definition

    def test_valid_config_validates(self, experiments):
        config_str = dedent(
            """
            [metrics.active_hours]
            select_expression = "1"
            data_source = "baseline"

            [metrics.active_hours.statistics.bootstrap_mean]

            [data_sources.baseline]
            from_expression = "mozdata.search.baseline"
            experiments_column_type = "simple"
            """
        )
        extern = Config(
            slug="cool_experiment",
            spec=self.spec,
            last_modified=datetime.datetime.now(),
        )
        definition = DefinitionConfig(
            slug="firefox_desktop",
            platform="firefox_desktop",
            spec=AnalysisSpec.from_dict(toml.loads(config_str)),
            last_modified=datetime.datetime.now(),
        )

        config_collection = ConfigCollection(
            configs=[extern], outcomes=[], defaults=[], definitions=[definition]
        )
        extern.validate(config_collection, experiments[0])

    def test_busted_config_fails(self, experiments):
        config = dedent(
            """\
            [metrics]
            weekly = ["bogus_metric"]

            [metrics.bogus_metric]
            select_expression = "SUM(fake_column)"
            data_source = "clients_daily"
            statistics = { bootstrap_mean = {} }
            """
        )
        spec = AnalysisSpec.from_dict(toml.loads(config))
        extern = Config(
            slug="bad_experiment",
            spec=spec,
            last_modified=datetime.datetime.now(),
        )
        config_collection = ConfigCollection([extern])
        with pytest.raises(DefinitionNotFound):
            extern.validate(config_collection, experiments[0])

    def test_valid_outcome_validates(self):
        config = dedent(
            """\
            friendly_name = "Fred"
            description = "Just your average paleolithic dad."

            [metrics.rocks_mined]
            select_expression = "COALESCE(SUM(pings_aggregated_by_this_row), 0)"
            data_source = "clients_daily"
            statistics = { bootstrap_mean = {} }
            friendly_name = "Rocks mined"
            description = "Number of rocks mined at the quarry"

            [data_sources.clients_daily]
            from_expression = "1"
            """
        )
        spec = OutcomeSpec.from_dict(toml.loads(config))
        extern = Outcome(
            slug="good_outcome",
            spec=spec,
            platform="firefox_desktop",
            commit_hash="0000000",
        )
        extern.validate(configs=ConfigCollection())

    def test_busted_outcome_fails(self):
        config = dedent(
            """\
            friendly_name = "Fred"
            description = "Just your average paleolithic dad."

            [metrics.rocks_mined]
            select_expression = "COALESCE(SUM(fake_column_whoop_whoop), 0)"
            data_source = "clients_daily"
            statistics = { bootstrap_mean = {} }
            friendly_name = "Rocks mined"
            description = "Number of rocks mined at the quarry"
            """
        )
        spec = OutcomeSpec.from_dict(toml.loads(config))
        extern = Outcome(
            slug="bogus_outcome",
            spec=spec,
            platform="firefox_desktop",
            commit_hash="0000000",
        )
        with pytest.raises(DefinitionNotFound):
            extern.validate(configs=ConfigCollection())

    def test_valid_default_config_validates(self):
        config_str = dedent(
            """
            [metrics.active_hours]
            select_expression = "1"
            data_source = "baseline"

            [metrics.active_hours.statistics.bootstrap_mean]

            [data_sources.baseline]
            from_expression = "mozdata.search.baseline"
            experiments_column_type = "simple"
            """
        )
        extern = Config(
            slug="cool_experiment",
            spec=self.spec,
            last_modified=datetime.datetime.now(),
        )
        definition = DefinitionConfig(
            slug="firefox_desktop",
            platform="firefox_desktop",
            spec=AnalysisSpec.from_dict(toml.loads(config_str)),
            last_modified=datetime.datetime.now(),
        )
        extern = DefaultConfig(
            slug="firefox_desktop",
            spec=self.spec,
            last_modified=datetime.datetime.now(),
        )
        extern.validate(configs=ConfigCollection(definitions=[definition]))

    def test_busted_default_config_fails(self):
        config = dedent(
            """\
            [metrics]
            weekly = ["bogus_metric"]

            [metrics.bogus_metric]
            select_expression = "SUM(fake_column)"
            data_source = "clients_daily"
            statistics = { bootstrap_mean = {} }
            """
        )
        spec = AnalysisSpec.from_dict(toml.loads(config))
        extern = DefaultConfig(
            slug="firefox_desktop",
            spec=spec,
            last_modified=datetime.datetime.now(),
        )
        with pytest.raises(DefinitionNotFound):
            extern.validate(configs=ConfigCollection())

    def test_merge_config_collection(self):
        config_str = dedent(
            """
            [metrics.active_hours]
            select_expression = "1"
            data_source = "baseline"

            [metrics.active_hours.statistics.bootstrap_mean]

            [data_sources.baseline]
            from_expression = "mozdata.search.baseline"
            experiments_column_type = "simple"
            """
        )
        extern = Config(
            slug="cool_experiment",
            spec=self.spec,
            last_modified=datetime.datetime.now(),
        )
        definition = DefinitionConfig(
            slug="firefox_desktop",
            platform="firefox_desktop",
            spec=AnalysisSpec.from_dict(toml.loads(config_str)),
            last_modified=datetime.datetime.now(),
        )
        config_collection_1 = ConfigCollection(
            configs=[extern], outcomes=[], defaults=[], definitions=[definition]
        )

        config_str = dedent(
            """
            [metrics.active_hours]
            select_expression = "1"
            data_source = "baseline"

            [metrics.active_hours.statistics.bootstrap_mean]

            [data_sources.baseline]
            from_expression = "mozdata.search.baseline"
            experiments_column_type = "simple"
            """
        )
        extern = Config(
            slug="cool_experiment_2",
            spec=self.spec,
            last_modified=datetime.datetime.now(),
        )
        definition = DefinitionConfig(
            slug="firefox_desktop",
            platform="firefox_desktop",
            spec=AnalysisSpec.from_dict(toml.loads(config_str)),
            last_modified=datetime.datetime.now(),
        )
        config_collection_2 = ConfigCollection(
            configs=[extern], outcomes=[], defaults=[], definitions=[definition]
        )

        config_collection_1.merge(config_collection_2)
        assert config_collection_1.configs[0].slug == "cool_experiment"
        assert config_collection_1.configs[1].slug == "cool_experiment_2"

    def test_merge_config_collection_override(self):
        config_str = dedent(
            """
            [metrics.active_hours]
            select_expression = "1"
            data_source = "baseline"

            [metrics.active_hours.statistics.bootstrap_mean]

            [data_sources.baseline]
            from_expression = "mozdata.search.baseline"
            experiments_column_type = "simple"
            """
        )
        extern = Config(
            slug="cool_experiment",
            spec=self.spec,
            last_modified=datetime.datetime.now(),
        )
        config_collection_1 = ConfigCollection(
            configs=[extern], outcomes=[], defaults=[], definitions=[]
        )

        config_str = dedent(
            """
            [metrics.active_hours]
            select_expression = "4"
            data_source = "baseline"

            [metrics.active_hours.statistics.bootstrap_mean]

            [metrics.unenroll]
            select_expression = "3"
            data_source = "baseline"

            [metrics.unenroll.statistics.bootstrap_mean]

            [data_sources.baseline]
            from_expression = "mozdata.search.baseline"
            experiments_column_type = "simple"
            """
        )
        extern = Config(
            slug="cool_experiment",
            spec=self.spec,
            last_modified=datetime.datetime.now(),
        )
        definition = DefinitionConfig(
            slug="firefox_desktop",
            platform="firefox_desktop",
            spec=AnalysisSpec.from_dict(toml.loads(config_str)),
            last_modified=datetime.datetime.now(),
        )
        config_collection_2 = ConfigCollection(
            configs=[extern], outcomes=[], defaults=[], definitions=[definition]
        )

        assert len(config_collection_1.definitions) == 0
        config_collection_1.merge(config_collection_2)

        assert len(config_collection_1.configs) == 1
        assert config_collection_1.configs[0].slug == "cool_experiment"

        assert [
            m
            for slug, m in config_collection_1.definitions[0].spec.metrics.definitions.items()
            if slug == "active_hours"
        ][0].select_expression == "4"
        assert [
            m
            for slug, m in config_collection_1.definitions[0].spec.metrics.definitions.items()
            if slug == "unenroll"
        ][0].select_expression == "3"

    def test_config_collection_from_subdir(self, local_tmp_repo):
        config_collection = ConfigCollection.from_github_repo(
            local_tmp_repo, path="metrics/jetstream"
        )
        assert len(config_collection.configs) > 0

    def test_configs_from_private_repo(self, local_tmp_repo):
        config_collection = ConfigCollection.from_github_repo(
            local_tmp_repo, path="metrics/jetstream", is_private=True
        )
        assert config_collection is not None
        assert config_collection.configs[0].spec.experiment.is_private

    def test_config_from_subdir(self, local_tmp_repo):
        nested_path = Path(local_tmp_repo) / "metrics" / "jetstream"
        config_collection = ConfigCollection.from_github_repo(nested_path)
        assert config_collection is not None

    def test_config_from_subdir_too_deep(self, local_tmp_repo):
        nested_path = Path(local_tmp_repo) / "metrics" / "jetstream"

        with pytest.raises(Exception):
            ConfigCollection.from_github_repo(nested_path, depth=1)

    def test_as_of_broken_commit(self, tmp_path):
        r = Repo.init(tmp_path)
        r.config_writer().set_value("user", "name", "test").release()
        r.config_writer().set_value("user", "email", "test@example.com").release()
        r.config_writer().set_value("commit", "gpgsign", "false").release()

        # check in broken file
        broken_config = dedent(
            """
            friendly_name = "Performance outcomes"
            description = "Outcomes related to performance"

            [metrics]
            weekly = ["speed"]
            overall = ["speed"]

            [metrics.speed]
            data_source = "main"
            select_expression = "1"

            [metrics.speed.statistics.bootstrap_mean]
            """
        )
        outcome_path = tmp_path / "jetstream" / "outcomes" / "firefox_desktop"
        outcome_path.mkdir(parents=True, exist_ok=True)
        (outcome_path / "performance.toml").write_text(broken_config)
        r.git.add(".")
        r.git.commit("-m", "commit", "--date", "Mon 20 Aug 2020 20:19:19 UTC")

        with pytest.raises(Exception):
            ConfigCollection.from_github_repo(tmp_path / "jetstream")

        with pytest.raises(Exception):
            ConfigCollection.from_github_repo(tmp_path / "jetstream").as_of(
                pytz.UTC.localize(datetime.datetime(2023, 5, 21))
            )

        # check in valid files
        shutil.copytree(TEST_DIR / "data", tmp_path, dirs_exist_ok=True)
        r.git.add(".")
        r.git.commit("-m", "commit", "--date", "Mon 25 Aug 2020 20:00:19 UTC")

        configs = ConfigCollection.from_github_repo(tmp_path / "jetstream")
        assert configs.outcomes is not None

        configs = configs.as_of(pytz.UTC.localize(datetime.datetime(2023, 5, 21)))
        assert configs.outcomes is not None

    def test_metric_level(self):
        config_str = dedent(
            """
            [metrics.active_hours]
            select_expression = "1"
            data_source = "baseline"
            owner = "me@example.com"
            category = "test"
            level = "silver"

            [metrics.other_metric]
            select_expression = "1"
            data_source = "baseline"
            owner = ["me@example.com", "you@example.com"]
            category = "test"
            level = "gold"

            [data_sources.baseline]
            from_expression = "mozdata.search.baseline"
            experiments_column_type = "simple"
            """
        )

        definition = DefinitionConfig(
            slug="firefox_desktop",
            platform="firefox_desktop",
            spec=AnalysisSpec.from_dict(toml.loads(config_str)),
            last_modified=datetime.datetime.now(),
        )
        config_collection = ConfigCollection(
            configs=[], outcomes=[], defaults=[], definitions=[definition]
        )

        metric_definition = config_collection.get_metric_definition(
            "active_hours", "firefox_desktop"
        )

        assert metric_definition.level == MetricLevel.SILVER
        assert metric_definition.owner == "me@example.com"
        assert metric_definition.category == "test"

        metric_definition = config_collection.get_metric_definition(
            "other_metric", "firefox_desktop"
        )

        assert metric_definition.level == MetricLevel.GOLD
        assert metric_definition.owner == ["me@example.com", "you@example.com"]
        assert metric_definition.category == "test"

    def test_invalid_metric_level(self):
        config_str = dedent(
            """
            [metrics.active_hours]
            select_expression = "1"
            data_source = "baseline"
            level = "invalid"

            [data_sources.baseline]
            from_expression = "mozdata.search.baseline"
            experiments_column_type = "simple"
            """
        )

        with pytest.raises(Exception):
            DefinitionConfig(
                slug="firefox_desktop",
                platform="firefox_desktop",
                spec=AnalysisSpec.from_dict(toml.loads(config_str)),
                last_modified=datetime.datetime.now(),
            )

    def test_valid_analysis_units(self):
        config_str = dedent(
            """
            [metrics.active_hours]
            select_expression = "1"
            data_source = "baseline"
            owner = "me@example.com"
            category = "test"
            analysis_units = ["profile_group_id"]

            [metrics.other_metric]
            select_expression = "1"
            data_source = "baseline"
            owner = ["me@example.com", "you@example.com"]
            category = "test"
            analysis_units = ["client_id", "profile_group_id"]

            [data_sources.baseline]
            from_expression = "mozdata.search.baseline"
            experiments_column_type = "simple"
            analysis_units = ["client_id", "profile_group_id"]
            """
        )

        definition = DefinitionConfig(
            slug="firefox_desktop",
            platform="firefox_desktop",
            spec=AnalysisSpec.from_dict(toml.loads(config_str)),
            last_modified=datetime.datetime.now(),
        )
        config_collection = ConfigCollection(
            configs=[], outcomes=[], defaults=[], definitions=[definition]
        )

        metric_definition = config_collection.get_metric_definition(
            "active_hours", "firefox_desktop"
        )
        assert metric_definition.analysis_units == [AnalysisUnit.PROFILE_GROUP]

        metric_definition = config_collection.get_metric_definition(
            "other_metric", "firefox_desktop"
        )
        assert metric_definition.analysis_units == [
            AnalysisUnit.CLIENT,
            AnalysisUnit.PROFILE_GROUP,
        ]

        data_source_definition = config_collection.get_data_source_definition(
            "baseline", "firefox_desktop"
        )
        assert data_source_definition.analysis_units == [
            AnalysisUnit.CLIENT,
            AnalysisUnit.PROFILE_GROUP,
        ]

    @pytest.mark.parametrize(
        "metric_units,ds_units",
        (
            ("analysis_units = 'client_id'", ""),
            ("analysis_units = ['invalid']", ""),
            ("", "analysis_units = 'client_id'"),
            ("", "analysis_units = ['invalid']"),
        ),
    )
    def test_invalid_analysis_units(self, metric_units, ds_units):
        config_str = dedent(
            f"""
            [metrics.active_hours]
            select_expression = "1"
            data_source = "baseline"
            {metric_units}

            [data_sources.baseline]
            from_expression = "mozdata.search.baseline"
            experiments_column_type = "simple"
            {ds_units}
            """
        )
        with pytest.raises(Exception):
            DefinitionConfig(
                slug="firefox_desktop",
                platform="firefox_desktop",
                spec=AnalysisSpec.from_dict(toml.loads(config_str)),
                last_modified=datetime.datetime.now(),
            )

    def test_data_source_joins(self):
        config_str = dedent(
            """
            [data_sources.events]
            from_expression = "mozdata.telemetry.events"
            experiments_column_type = "simple"

            [data_sources.metrics]
            from_expression = "mozdata.telemetry.metrics"
            experiments_column_type = "simple"

            [data_sources.baseline]
            from_expression = "mozdata.telemetry.baseline"
            experiments_column_type = "simple"

            [data_sources.baseline.joins.metrics]
            on_expression = "metrics.client_id = baseline.client_id"
            relationship = "many_to_one"

            [data_sources.baseline.joins.events]
            """
        )

        definition = DefinitionConfig(
            slug="firefox_desktop",
            platform="firefox_desktop",
            spec=AnalysisSpec.from_dict(toml.loads(config_str)),
            last_modified=datetime.datetime.now(),
        )
        config_collection = ConfigCollection(
            configs=[], outcomes=[], defaults=[], definitions=[definition]
        )

        data_source = config_collection.get_data_source_definition(
            "baseline", "firefox_desktop"
        ).resolve(definition.spec, None, config_collection)

        assert len(data_source.joins) == 2
        assert data_source.joins[0].data_source.name == "metrics"
        assert data_source.joins[0].on_expression == "metrics.client_id = baseline.client_id"
        assert data_source.joins[0].relationship == DataSourceJoinRelationship.MANY_TO_ONE

        assert data_source.joins[1].data_source.name == "events"

    def test_data_source_joins_invalid(self):
        config_str = dedent(
            """
            [data_sources.baseline]
            from_expression = "mozdata.telemetry.baseline"
            experiments_column_type = "simple"

            [data_sources.baseline.joins.non_existing]
            relationship = "many_to_one"
            """
        )

        definition = DefinitionConfig(
            slug="firefox_desktop",
            platform="firefox_desktop",
            spec=AnalysisSpec.from_dict(toml.loads(config_str)),
            last_modified=datetime.datetime.now(),
        )
        config_collection = ConfigCollection(
            configs=[], outcomes=[], defaults=[], definitions=[definition]
        )

        with pytest.raises(Exception):
            config_collection.get_data_source_definition("baseline", "firefox_desktop").resolve(
                definition.spec, None, config_collection
            )

    def test_data_source_joins_circular_dependency(self):
        config_str = dedent(
            """
            [data_sources.baseline]
            from_expression = "mozdata.telemetry.baseline"
            experiments_column_type = "simple"

            [data_sources.metrics]
            from_expression = "mozdata.telemetry.metrics"
            experiments_column_type = "simple"

            [data_sources.baseline.joins.metrics]

            [data_sources.metrics.joins.baseline]
            """
        )

        definition = DefinitionConfig(
            slug="firefox_desktop",
            platform="firefox_desktop",
            spec=AnalysisSpec.from_dict(toml.loads(config_str)),
            last_modified=datetime.datetime.now(),
        )
        config_collection = ConfigCollection(
            configs=[], outcomes=[], defaults=[], definitions=[definition]
        )

        with pytest.raises(Exception):
            config_collection.get_data_source_definition("baseline", "firefox_desktop").resolve(
                definition.spec, None, config_collection
            )

    def test_invalid_wildcard_in_data_source(self):
        # needs to be [data_sources.'baseline_*']
        config_str = dedent(
            """
            [data_sources.baseline_*]
            from_expression = "mozdata.search.baseline"
            experiments_column_type = "simple"
            """
        )

        with pytest.raises(ValueError):
            DefinitionConfig(
                slug="firefox_desktop",
                platform="firefox_desktop",
                spec=AnalysisSpec.from_dict(toml.loads(config_str)),
                last_modified=datetime.datetime.now(),
            )

    def test_merge_with_wildcards(self):
        config_str = dedent(
            """
            [metrics.test_metric]
            select_expression = 1
            category = "test"

            [data_sources.baseline]
            from_expression = "mozdata.search.baseline"
            experiments_column_type = "simple"
            """
        )

        definition = DefinitionConfig(
            slug="firefox_desktop",
            platform="firefox_desktop",
            spec=AnalysisSpec.from_dict(toml.loads(config_str)),
            last_modified=datetime.datetime.now(),
        )
        config_collection_1 = ConfigCollection(
            configs=[], outcomes=[], defaults=[], definitions=[definition]
        )

        config_str = dedent(
            """
            [data_sources.'*']
            experiments_column_type = "none"

            [metrics.'test_*']
            data_source = "baseline"
            category = "test_overwrite_first"

            [metrics.'test_*'.statistics.bootstrap_mean]

            [metrics.'test_me*']
            category = "test_overwrite_second"
            """
        )
        definition = DefinitionConfig(
            slug="firefox_desktop",
            platform="firefox_desktop",
            spec=AnalysisSpec.from_dict(toml.loads(config_str)),
            last_modified=datetime.datetime.now(),
        )
        config_collection_2 = ConfigCollection(
            configs=[], outcomes=[], defaults=[], definitions=[definition]
        )

        config_collection_1.get_metric_definition(
            "test_metric", "firefox_desktop"
        ).data_source is None
        assert (
            config_collection_1.get_data_source_definition(
                "baseline", "firefox_desktop"
            ).experiments_column_type
            == "simple"
        )

        config_collection_1.merge(config_collection_2)
        assert (
            config_collection_1.get_data_source_definition("baseline", "firefox_desktop").name
            == "baseline"
        )
        assert (
            config_collection_1.get_data_source_definition(
                "baseline", "firefox_desktop"
            ).experiments_column_type
            == "none"
        )
        assert config_collection_1.get_data_source_definition("*", "firefox_desktop") is None
        assert (
            config_collection_1.get_metric_definition(
                "test_metric", "firefox_desktop"
            ).data_source.name
            == "baseline"
        )
        assert (
            "bootstrap_mean"
            in config_collection_1.get_metric_definition(
                "test_metric", "firefox_desktop"
            ).statistics
        )
        assert (
            config_collection_1.get_metric_definition("test_metric", "firefox_desktop").category
            == "test_overwrite_second"
        )

    def test_merge_with_wildcards_invalid(self):
        config_str = dedent(
            """
            [data_sources.baseline]
            from_expression = "mozdata.search.baseline"
            experiments_column_type = "simple"
            """
        )

        definition = DefinitionConfig(
            slug="firefox_desktop",
            platform="firefox_desktop",
            spec=AnalysisSpec.from_dict(toml.loads(config_str)),
            last_modified=datetime.datetime.now(),
        )
        config_collection_1 = ConfigCollection(
            configs=[], outcomes=[], defaults=[], definitions=[definition]
        )

        config_str = dedent(
            """
            [data_sources.'invalid_*']
            experiments_column_type = "none"
            """
        )
        definition = DefinitionConfig(
            slug="firefox_desktop",
            platform="firefox_desktop",
            spec=AnalysisSpec.from_dict(toml.loads(config_str)),
            last_modified=datetime.datetime.now(),
        )
        config_collection_2 = ConfigCollection(
            configs=[], outcomes=[], defaults=[], definitions=[definition]
        )

        config_collection_1.merge(config_collection_2)
        assert (
            config_collection_1.get_data_source_definition("baseline", "firefox_desktop").name
            == "baseline"
        )
        assert (
            config_collection_1.get_data_source_definition(
                "baseline", "firefox_desktop"
            ).experiments_column_type
            == "simple"
        )
        assert (
            config_collection_1.get_data_source_definition("invalid_*", "firefox_desktop") is None
        )

    def test_get_segments_for_app_with_toml(self):
        config_str_desktop = """
        [segments.my_cool_segment]
        data_source = "my_cool_data_source"
        select_expression = "{{agg_any('1')}}"

        [segments.another_segment]
        data_source = "another_data_source"
        select_expression = "{{agg_sum('2')}}"

        [segments.data_sources.my_cool_data_source]
        from_expression = "max(attr_source is not null)"

        [segments.data_sources.another_data_source]
        from_expression = "max(attr_source is null)"
        """

        config_str_fenix = """
        [segments.fenix_segment]
        data_source = "fenix_data_source"
        select_expression = "{{agg_count('3')}}"

        [segments.data_sources.fenix_data_source]
        from_expression = "min(attr_source is not null)"
        """

        definition_desktop = DefinitionConfig(
            slug="firefox_desktop",
            platform="firefox_desktop",
            spec=AnalysisSpec.from_dict(toml.loads(config_str_desktop)),
            last_modified=datetime.datetime.now(),
        )

        definition_fenix = DefinitionConfig(
            slug="fenix",
            platform="fenix",
            spec=AnalysisSpec.from_dict(toml.loads(config_str_fenix)),
            last_modified=datetime.datetime.now(),
        )

        config_collection = ConfigCollection(
            configs=[],
            outcomes=[],
            defaults=[],
            definitions=[definition_desktop, definition_fenix],
        )

        segments_desktop = config_collection.get_segments_for_app("firefox_desktop")
        assert len(segments_desktop) == 2

        segment_slugs_desktop = [seg.name for seg in segments_desktop]
        assert "my_cool_segment" in segment_slugs_desktop
        assert "another_segment" in segment_slugs_desktop

        segments_fenix = config_collection.get_segments_for_app("fenix")
        assert len(segments_fenix) == 1

        segment_slugs_fenix = [seg.name for seg in segments_fenix]
        assert "fenix_segment" in segment_slugs_fenix

    def test_local_config_collection_from_local_path(self):
        config_collection = LocalConfigCollection.from_local_path(TEST_DIR / "data")
        segments = [s.name for s in config_collection.get_segments_for_app("firefox_desktop")]
        assert "regular_users_v3" in segments
        assert len(segments) == 1

        outcomes = [o.slug for o in config_collection.outcomes]
        assert len(outcomes) == 0

        test_metric = config_collection.get_metric_definition("active_hours", "firefox_desktop")
        assert test_metric.name == "active_hours"
        assert test_metric.select_expression == '{{agg_sum("active_hours_sum")}}'
        assert test_metric.data_source.name == "clients_daily"

        cc_jetstream = LocalConfigCollection.from_local_path(TEST_DIR / "data" / "jetstream")
        config_collection.merge(cc_jetstream)

        outcomes = [o.slug for o in config_collection.outcomes]
        assert len(outcomes) == 4



from textwrap import dedent
from typing import TYPE_CHECKING, Any, Dict, Optional

import attr
import jinja2
from jinja2 import StrictUndefined

if TYPE_CHECKING:
    from .config import ConfigCollection
    from .analysis import AnalysisSpec
    from .experiment import ExperimentConfiguration

from . import AnalysisUnit
from .errors import DefinitionNotFound
from .util import converter


@attr.s(frozen=True, slots=True)
class SegmentDataSource:
    """Represents a table or view, from which segments may be defined.
    ``window_start`` and ``window_end`` define the window of data used
    to determine whether each client fits a segment. Ideally this
    window ends at/before the moment of enrollment, so that user's
    branches can't bias the segment assignment. ``window_start`` and
    ``window_end`` are integers, representing the number
    of days before or after enrollment.
    Args:
        name (str): Name for the Data Source. Should be unique to avoid
            confusion.
        from_expression (str): FROM expression - often just a fully-qualified
            table name. Sometimes a subquery. May contain the string
            ``{dataset}`` which will be replaced with an app-specific
            dataset for Glean apps. If the expression is templated
            on dataset, default_dataset is mandatory.
        window_start (int, optional): See above.
        window_end (int, optional): See above.
        client_id_column (str, optional): Name of the column that
            contains the ``client_id`` (join key). Defaults to
            'client_id'.
        submission_date_column (str, optional): Name of the column
            that contains the submission date (as a date, not
            timestamp). Defaults to 'submission_date'.
        default_dataset (str, optional): The value to use for
            `{dataset}` in from_expression if a value is not provided
            at runtime. Mandatory if from_expression contains a
            `{dataset}` parameter.
        group_id_column (str, optional): Name of the column that
            contains the ``profile_group_id`` (join key). Defaults to
            'profile_group_id'.
        glean_client_id_column (str, optional): Name of the column that
            contains the *glean* telemetry ``client_id`` (join key).
            This is also used to specify that the data source supports glean.
        legacy_client_id_column (str, optional): Name of the column that
            contains the *legacy* telemetry ``client_id`` (join key).
            This is also used to specify that the data source supports legacy.
    """

    name = attr.ib(validator=attr.validators.instance_of(str))
    from_expression = attr.ib(validator=attr.validators.instance_of(str))
    window_start = attr.ib(default=0, type=int)
    window_end = attr.ib(default=0, type=int)
    client_id_column = attr.ib(default=AnalysisUnit.CLIENT.value, type=str)
    submission_date_column = attr.ib(default="submission_date", type=str)
    default_dataset = attr.ib(default=None, type=Optional[str])
    group_id_column = attr.ib(default=AnalysisUnit.PROFILE_GROUP.value, type=str)
    glean_client_id_column = attr.ib(default=None, type=str)
    legacy_client_id_column = attr.ib(default=None, type=str)


@attr.s(frozen=True, slots=True)
class Segment:
    """Represents an experiment Segment.
    Args:
        name (str): The segment's name; will be a column name.
        data_source (SegmentDataSource): Data source that provides
            the columns referenced in ``select_expression``.
        select_expression (str): A SQL select expression that includes
            an aggregation function (we ``GROUP BY client_id``).
            Returns a non-NULL ``BOOL``: ``True`` if the user is in the
            segment, ``False`` otherwise.
        friendly_name (str): A human-readable dashboard title for this segment
        description (str): A paragraph of Markdown-formatted text describing
            the segment in more detail, to be shown on dashboards
    """

    name = attr.ib(type=str)
    data_source = attr.ib(validator=attr.validators.instance_of(SegmentDataSource))
    select_expression = attr.ib(type=str)
    friendly_name = attr.ib(type=Optional[str], default=None)
    description = attr.ib(type=Optional[str], default=None)


@attr.s(auto_attribs=True)
class SegmentReference:
    name: str

    def resolve(
        self,
        spec: "AnalysisSpec",
        conf: "ExperimentConfiguration",
        configs: "ConfigCollection",
    ) -> Segment:
        if self.name in spec.segments.definitions:
            return spec.segments.definitions[self.name].resolve(spec, conf, configs)
        segment_definition = configs.get_segment_definition(self.name, conf.app_name)
        if segment_definition is None:
            raise DefinitionNotFound(f"Could not find definition for segment '{self.name}'")
        return segment_definition.resolve(spec, conf, configs)


converter.register_structure_hook(SegmentReference, lambda obj, _type: SegmentReference(name=obj))


@attr.s(auto_attribs=True)
class SegmentDataSourceDefinition:
    name: str
    from_expression: str
    window_start: int = 0
    window_end: int = 0
    client_id_column: Optional[str] = AnalysisUnit.CLIENT.value
    submission_date_column: Optional[str] = "submission_date"
    default_dataset: Optional[str] = None
    group_id_column: Optional[str] = AnalysisUnit.PROFILE_GROUP.value
    glean_client_id_column: Optional[str] = None
    legacy_client_id_column: Optional[str] = None

    def resolve(
        self,
        spec: "AnalysisSpec",
        conf: "ExperimentConfiguration",
        _configs: "ConfigCollection",
    ) -> SegmentDataSource:
        env = jinja2.Environment(autoescape=False, undefined=StrictUndefined)
        from_expression = env.from_string(self.from_expression).render(experiment=conf)
        kwargs: Dict[str, Any] = {
            "name": self.name,
            "from_expression": from_expression,
            "window_start": self.window_start,
            "window_end": self.window_end,
        }
        for k in (
            "client_id_column",
            "submission_date_column",
            "group_id_column",
            "glean_client_id_column",
            "legacy_client_id_column",
        ):
            v = getattr(self, k)
            if v:
                kwargs[k] = v
        return SegmentDataSource(**kwargs)


@attr.s(auto_attribs=True)
class SegmentDataSourceReference:
    name: str

    def resolve(
        self,
        spec: "AnalysisSpec",
        conf: "ExperimentConfiguration",
        configs: "ConfigCollection",
    ) -> SegmentDataSource:
        if self.name in spec.segments.data_sources:
            return spec.segments.data_sources[self.name].resolve(spec, conf, configs)
        segment_definition = configs.get_segment_data_source_definition(self.name, conf.app_name)
        if segment_definition is None:
            raise DefinitionNotFound(
                f"Could not find definition for segment data source '{self.name}'"
            )
        return segment_definition.resolve(spec, conf, configs)


converter.register_structure_hook(
    SegmentDataSourceReference, lambda obj, _type: SegmentDataSourceReference(name=obj)
)


@attr.s(auto_attribs=True)
class SegmentDefinition:
    name: str
    data_source: SegmentDataSourceReference
    select_expression: str
    friendly_name: Optional[str] = None
    description: Optional[str] = None

    def resolve(
        self,
        spec: "AnalysisSpec",
        conf: "ExperimentConfiguration",
        configs: "ConfigCollection",
    ) -> Segment:
        data_source = self.data_source.resolve(spec, conf, configs)

        return Segment(
            name=self.name,
            data_source=data_source,
            select_expression=configs.get_env().from_string(self.select_expression).render(),
            friendly_name=(
                dedent(self.friendly_name) if self.friendly_name else self.friendly_name
            ),
            description=(dedent(self.description) if self.description else self.description),
        )


@attr.s(auto_attribs=True)
class SegmentsSpec:
    definitions: Dict[str, SegmentDefinition] = attr.Factory(dict)
    data_sources: Dict[str, SegmentDataSourceDefinition] = attr.Factory(dict)

    @classmethod
    def from_dict(cls, d: dict) -> "SegmentsSpec":
        data_sources = {
            k: converter.structure(
                {"name": k, **dict((kk.lower(), vv) for kk, vv in v.items())},
                SegmentDataSourceDefinition,
            )
            for k, v in d.pop("data_sources", {}).items()
        }
        definitions = {
            k: converter.structure(
                {"name": k, **dict((kk.lower(), vv) for kk, vv in v.items())},
                SegmentDefinition,
            )
            for k, v in d.items()
        }
        return cls(definitions, data_sources)

    def merge(self, other: "SegmentsSpec"):
        """
        Merge another segments spec into the current one.
        The `other` SegmentsSpec overwrites existing keys.
        """
        self.data_sources.update(other.data_sources)
        self.definitions.update(other.definitions)


converter.register_structure_hook(SegmentsSpec, lambda obj, _type: SegmentsSpec.from_dict(obj))



import os
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

from .metric import MetricDefinition

if TYPE_CHECKING:
    from .config import ConfigCollection

FILE_PATH = Path(os.path.dirname(__file__))
METRICS_QUERY = FILE_PATH / "templates" / "metrics_query.sql"
DATA_SOURCE_QUERY = FILE_PATH / "templates" / "data_source_query.sql"
DATA_SOURCE_MACROS = FILE_PATH / "templates" / "data_source_macros.j2"


def generate_metrics_sql(
    config_collection: "ConfigCollection",
    metrics: List[str],
    platform: str,
    group_by: Union[List[str], Dict[str, str]] = [],
    where: Optional[str] = None,
    group_by_client_id: bool = True,
    group_by_submission_date: bool = True,
) -> str:
    """Generates a SQL query for metrics and specified parameters."""
    metric_definitions: List[MetricDefinition] = []
    for slug in metrics:
        definition = config_collection.get_metric_definition(slug, platform)

        if definition is None:
            raise ValueError(f"No definition for metric {slug} on platform {platform} found.")

        metric_definitions.append(definition)

    metrics_per_data_source: Dict[str, Any] = {}
    for metric in metric_definitions:
        if metric.select_expression is None:
            raise ValueError(f"No definition for metric {metric.name}")

        metric.select_expression = (
            config_collection.get_env().from_string(metric.select_expression).render()
        )

        if metric.data_source is None:
            raise ValueError(f"No data source for metric {metric.name}")

        if metric.data_source.name in metrics_per_data_source:
            metrics_per_data_source[metric.data_source.name]["metrics"].append(metric)
        else:
            data_source = config_collection.get_data_source_definition(
                metric.data_source.name, platform
            )

            if data_source is None:
                raise ValueError(f"No valid data source definition found for metric {metric.name}")

            # default parameters need to be set explicitly otherwise they'll be None
            data_source.client_id_column = data_source.client_id_column or "client_id"
            data_source.submission_date_column = (
                data_source.submission_date_column or "submission_date"
            )

            metrics_per_data_source[metric.data_source.name] = {
                "data_source": data_source,
                "metrics": [metric],
            }

    # group by should be a dictionary with the key being the alias and
    # the value the potentially nested field;
    # it can also be specified as list if all fields are top-level fields that don't need an alias
    if isinstance(group_by, list):
        group_by = {g: g for g in group_by}

    macros_template = DATA_SOURCE_MACROS.read_text()
    template = METRICS_QUERY.read_text()

    # using `from_string()` in Jinja doens't support include statements, so
    # substituting them here manually
    template = template.replace("{% include 'data_source_macros.j2' %}", macros_template)
    return (
        config_collection.get_env()
        .from_string(template)
        .render(
            **{
                "metrics_per_data_source": metrics_per_data_source,
                "where": where,
                "group_by": group_by,
                "group_by_client_id": group_by_client_id,
                "group_by_submission_date": group_by_submission_date,
                "data_sources": {
                    slug: data_source
                    for definition in config_collection.definitions
                    for slug, data_source in definition.spec.data_sources.definitions.items()
                    if platform == definition.platform
                },
                "select_fields": True,
                "ignore_joins": False,
            }
        )
    )


def generate_data_source_sql(
    config_collection: "ConfigCollection",
    data_source: str,
    platform: str,
    where: Optional[str] = None,
    select_fields: bool = True,
    ignore_joins: bool = False,
) -> str:
    """Generates a SQL query for the specified data source."""
    template = DATA_SOURCE_QUERY.read_text()
    macros_template = DATA_SOURCE_MACROS.read_text()

    # using `from_string()` in Jinja doens't support include statements, so
    # substituting them here manually
    template = template.replace("{% include 'data_source_macros.j2' %}", macros_template)
    data_source_definition = config_collection.get_data_source_definition(data_source, platform)

    if data_source_definition is None:
        raise ValueError(f"No valid data source definition found for {data_source}")

    # default parameters need to be set explicitly otherwise they'll be None
    data_source_definition.client_id_column = data_source_definition.client_id_column or "client_id"
    data_source_definition.submission_date_column = (
        data_source_definition.submission_date_column or "submission_date"
    )
    return (
        config_collection.get_env()
        .from_string(template)
        .render(
            **{
                "data_source": data_source_definition,
                "data_sources": {
                    slug: data_source
                    for definition in config_collection.definitions
                    for slug, data_source in definition.spec.data_sources.definitions.items()
                    if platform == definition.platform
                },
                "where": where,
                "select_fields": select_fields,
                "ignore_joins": ignore_joins,
            }
        )
    )



import datetime as dt
from pathlib import Path
from textwrap import dedent

import pytest
import pytz
import toml
from cattrs.errors import ClassValidationError
from mozilla_nimbus_schemas import RandomizationUnit

from metric_config_parser import AnalysisUnit
from metric_config_parser.analysis import AnalysisSpec
from metric_config_parser.errors import NoEndDateException
from metric_config_parser.experiment import Branch, BucketConfig, Experiment
from metric_config_parser.metric import AnalysisPeriod
from metric_config_parser.segment import Segment

TEST_DIR = Path(__file__).parent
DEFAULT_METRICS_CONFIG = TEST_DIR / "data" / "jetstream" / "defaults" / "firefox_desktop.toml"


class TestExperimentSpec:
    def test_null_query(self, experiments, config_collection):
        spec = AnalysisSpec.from_dict({})
        cfg = spec.resolve(experiments[0], config_collection)
        assert cfg.experiment.enrollment_query is None

    def test_trivial_query(self, experiments, config_collection):
        conf = dedent(
            """
            [experiment]
            enrollment_query = "SELECT 1"
            """
        )
        spec = AnalysisSpec.from_dict(toml.loads(conf))
        cfg = spec.resolve(experiments[0], config_collection)
        assert cfg.experiment.enrollment_query == "SELECT 1"

    def test_template_query(self, experiments, config_collection):
        conf = dedent(
            """
            [experiment]
            enrollment_query = "SELECT 1 FROM foo WHERE slug = '{{experiment.experimenter_slug}}'"
            """
        )
        spec = AnalysisSpec.from_dict(toml.loads(conf))
        cfg = spec.resolve(experiments[0], config_collection)
        assert cfg.experiment.enrollment_query == "SELECT 1 FROM foo WHERE slug = 'test_slug'"

    def test_silly_query(self, experiments, config_collection):
        conf = dedent(
            """
            [experiment]
            enrollment_query = "{{experiment.enrollment_query}}"  # whoa
            """
        )
        spec = AnalysisSpec.from_dict(toml.loads(conf))
        with pytest.raises(ValueError):
            spec.resolve(experiments[0], config_collection)

    def test_control_branch(self, experiments, config_collection):
        trivial = AnalysisSpec().resolve(experiments[0], config_collection)
        assert trivial.experiment.reference_branch == "b"

        conf = dedent(
            """
            [experiment]
            reference_branch = "a"
            """
        )
        spec = AnalysisSpec.from_dict(toml.loads(conf))
        configured = spec.resolve(experiments[0], config_collection)
        assert configured.experiment.reference_branch == "a"

    def test_recognizes_segments(self, experiments, config_collection):
        conf = dedent(
            """
            [experiment]
            segments = ["regular_users_v3"]
            """
        )
        spec = AnalysisSpec.from_dict(toml.loads(conf))
        configured = spec.resolve(experiments[0], config_collection)
        assert isinstance(configured.experiment.segments[0], Segment)

    def test_segment_definitions(self, experiments, config_collection):
        conf = dedent(
            """
            [experiment]
            segments = ["regular_users_v3", "my_cool_segment"]

            [segments.my_cool_segment]
            Data_Source = "my_cool_data_source"
            Select_Expression = "{{agg_any('1')}}"

            [segments.data_sources.my_cool_data_source]
            from_expression = "(SELECT 1 WHERE submission_date BETWEEN {{experiment.start_date_str}} AND {{experiment.last_enrollment_date_str}})"
            """  # noqa
        )

        spec = AnalysisSpec.from_dict(toml.loads(conf))
        configured = spec.resolve(experiments[0], config_collection)

        assert len(configured.experiment.segments) == 2

        for segment in configured.experiment.segments:
            assert isinstance(segment, Segment)

        assert configured.experiment.segments[0].name == "regular_users_v3"
        assert configured.experiment.segments[1].name == "my_cool_segment"

        assert "agg_any" not in configured.experiment.segments[1].select_expression
        assert "1970" not in configured.experiment.segments[1].data_source.from_expression
        assert "{{" not in configured.experiment.segments[1].data_source.from_expression
        assert "2019-12-01" in configured.experiment.segments[1].data_source.from_expression

    def test_end_date_str(self, experiments, config_collection):
        conf = dedent(
            """
            [experiment]
            segments = ["my_cool_segment"]

            [segments.my_cool_segment]
            Data_Source = "my_cool_data_source"
            Select_Expression = "{{agg_any('1')}}"

            [segments.data_sources.my_cool_data_source]
            from_expression = "(SELECT 1 WHERE submission_date BETWEEN {{experiment.start_date_str}} AND {{experiment.end_date_str}})"
            """  # noqa
        )

        spec = AnalysisSpec.from_dict(toml.loads(conf))
        configured = spec.resolve(experiments[0], config_collection)

        assert len(configured.experiment.segments) == 1

        segment = configured.experiment.segments[0]
        assert isinstance(segment, Segment)

        assert segment.name == "my_cool_segment"

        assert "agg_any" not in segment.select_expression
        assert "1970" not in segment.data_source.from_expression
        assert "{{" not in segment.data_source.from_expression
        assert "2019-12-01" in segment.data_source.from_expression
        assert "2020-03-01" in segment.data_source.from_expression

        # Fails when `end_date=None`.
        with pytest.raises(NoEndDateException):
            spec = AnalysisSpec.from_dict(toml.loads(conf))
            configured = spec.resolve(experiments[8], config_collection)

        # Succeeds when `end_date=None` but it's not referenced; note
        # `last_enrollment_date_str` below.
        conf = dedent(
            """
            [experiment]
            segments = ["my_cool_segment"]

            [segments.my_cool_segment]
            Data_Source = "my_cool_data_source"
            Select_Expression = "{{agg_any('1')}}"

            [segments.data_sources.my_cool_data_source]
            from_expression = "(SELECT 1 WHERE submission_date BETWEEN {{experiment.start_date_str}} AND {{experiment.last_enrollment_date_str}})"
            """  # noqa
        )

        spec = AnalysisSpec.from_dict(toml.loads(conf))
        configured = spec.resolve(experiments[8], config_collection)

    def test_pre_treatment_config(self, experiments, config_collection):
        config_str = dedent(
            """
            [metrics]
            weekly = ["spam"]

            [metrics.spam]
            data_source = "main"
            select_expression = "1"

            [metrics.spam.statistics.bootstrap_mean]
            num_samples = 10
            pre_treatments = [
                {name = "remove_nulls"},
                {name = "log", base = 20.0},
                {name = "censor_highest_values", fraction = 0.9}
            ]
            """
        )

        spec = AnalysisSpec.from_dict(toml.loads(config_str))
        cfg = spec.resolve(experiments[0], config_collection)
        pre_treatments = [m for m in cfg.metrics[AnalysisPeriod.WEEK] if m.metric.name == "spam"][
            0
        ].pre_treatments

        assert len(pre_treatments) == 3
        assert pre_treatments[0].name == "remove_nulls"
        assert pre_treatments[1].name == "log"
        assert pre_treatments[2].name == "censor_highest_values"

        assert pre_treatments[1].args["base"] == 20.0
        assert pre_treatments[2].args["fraction"] == 0.9

    def test_pre_treatment_config_multiple_periods(self, experiments, config_collection):
        config_str = dedent(
            """
            [metrics]
            weekly = ["spam"]
            28_day = ["spam"]
            overall = ["spam"]

            [metrics.spam]
            data_source = "main"
            select_expression = "1"

            [metrics.spam.statistics.binomial]
            pre_treatments = ["remove_nulls"]
            """
        )

        spec = AnalysisSpec.from_dict(toml.loads(config_str))
        cfg = spec.resolve(experiments[0], config_collection)
        pre_treatments = [m for m in cfg.metrics[AnalysisPeriod.WEEK] if m.metric.name == "spam"][
            0
        ].pre_treatments

        assert len(pre_treatments) == 1
        assert pre_treatments[0].name == "remove_nulls"

        overall_pre_treatments = [
            m for m in cfg.metrics[AnalysisPeriod.OVERALL] if m.metric.name == "spam"
        ][0].pre_treatments

        assert len(overall_pre_treatments) == 1
        assert overall_pre_treatments[0].name == "remove_nulls"

    def test_preenrollment(self, experiments, config_collection):
        config_str = dedent(
            """
            [metrics]
            preenrollment_days28 = ["spam"]
            preenrollment_weekly = ["spam"]

            [metrics.spam]
            data_source = "main"
            select_expression = "1"

            [metrics.spam.statistics.binomial]
            """
        )

        spec = AnalysisSpec.from_dict(toml.loads(config_str))
        cfg = spec.resolve(experiments[0], config_collection)
        week_metrics = [
            m for m in cfg.metrics[AnalysisPeriod.PREENROLLMENT_WEEK] if m.metric.name == "spam"
        ]

        assert len(week_metrics) == 1
        assert week_metrics[0].metric.name == "spam"

        days28_metrics = [
            m for m in cfg.metrics[AnalysisPeriod.PREENROLLMENT_DAYS_28] if m.metric.name == "spam"
        ]

        assert len(days28_metrics) == 1
        assert days28_metrics[0].metric.name == "spam"

    def test_segment_resolution(self, experiments, config_collection):
        spec = AnalysisSpec()

        cfg = spec.resolve(experiments[9], config_collection)

        assert len(cfg.experiment.segments) == 1

        segment = cfg.experiment.segments[0]

        assert segment.name == "regular_users_v3"
        assert segment.data_source.name == "my_cool_data_source"
        expected_expression = """(SELECT 1 WHERE submission_date BETWEEN 2019-12-01
    AND 2019-12-08)"""
        assert expected_expression == segment.data_source.from_expression.strip()

    def test_segment_definition_and_experiment_segment(self, experiments, config_collection):
        config_str = dedent(
            """
            [experiment]
            segments = ["test_segment"]

            [segments.test_segment]
            Data_Source = "test_data_source"
            Select_Expression = "{{agg_any('1')}}"

            [segments.data_sources.test_data_source]
            from_expression = "(SELECT 1 WHERE submission_date BETWEEN {{experiment.start_date_str}} AND {{experiment.last_enrollment_date_str}})"
            """  # noqa
        )
        spec = AnalysisSpec.from_dict(toml.loads(config_str))
        cfg = spec.resolve(experiments[9], config_collection)

        assert len(cfg.experiment.segments) == 2

        custom_segment = [seg for seg in cfg.experiment.segments if seg.name == "test_segment"]

        assert len(custom_segment) == 1
        assert custom_segment[0].name == "test_segment"
        assert custom_segment[0].data_source.name == "test_data_source"
        expected_expression = (
            """(SELECT 1 WHERE submission_date BETWEEN 2019-12-01 AND 2019-12-08)"""
        )
        assert expected_expression == custom_segment[0].data_source.from_expression.strip()

        experiment_segments = [
            seg for seg in cfg.experiment.segments if seg.name == "regular_users_v3"
        ]

        assert len(experiment_segments) == 1
        assert experiment_segments[0].name == "regular_users_v3"
        assert experiment_segments[0].data_source.name == "my_cool_data_source"
        expected_expression = """(SELECT 1 WHERE submission_date BETWEEN 2019-12-01
    AND 2019-12-08)"""
        assert expected_expression == experiment_segments[0].data_source.from_expression.strip()

    def test_segment_duplication(self, experiments, config_collection):
        config_str = dedent(
            """
            [experiment]
            segments = ["regular_users_v3"]

            [segments.regular_users_v3]
            data_source = "my_cool_data_source"
            select_expression = "{{agg_any('1')}}"

            [segments.data_sources.my_cool_data_source]
            from_expression = '''
                (SELECT 1 WHERE submission_date BETWEEN {{experiment.start_date_str}}
                AND {{experiment.last_enrollment_date_str}})
            '''
            """
        )

        spec = AnalysisSpec.from_dict(toml.loads(config_str))
        cfg = spec.resolve(experiments[9], config_collection)

        assert len(cfg.experiment.segments) == 1
        assert cfg.experiment.segments[0].name == "regular_users_v3"
        assert cfg.experiment.segments[0].data_source.name == "my_cool_data_source"
        expected_expression = """(SELECT 1 WHERE submission_date BETWEEN 2019-12-01
    AND 2019-12-08)"""
        assert expected_expression == cfg.experiment.segments[0].data_source.from_expression.strip()


class TestExperimentConf:
    def test_bad_dates(self, experiments):
        conf = dedent(
            """
            [experiment]
            end_date = "Christmas"
            """
        )
        with pytest.raises(ClassValidationError):
            AnalysisSpec.from_dict(toml.loads(conf))

        conf = dedent(
            """
            [experiment]
            start_date = "My birthday"
            """
        )
        with pytest.raises(ClassValidationError):
            AnalysisSpec.from_dict(toml.loads(conf))

    def test_good_end_date(self, experiments, config_collection):
        conf = dedent(
            """
            [experiment]
            end_date = "2020-12-31"
            """
        )
        spec = AnalysisSpec.from_dict(toml.loads(conf))
        live_experiment = [x for x in experiments if x.status == "Live"][0]
        cfg = spec.resolve(live_experiment, config_collection)
        assert cfg.experiment.end_date == dt.datetime(2020, 12, 31, tzinfo=pytz.utc)
        assert cfg.experiment.status == "Complete"

    def test_good_start_date(self, experiments, config_collection):
        conf = dedent(
            """
            [experiment]
            start_date = "2020-12-31"
            end_date = "2021-02-01"
            """
        )
        spec = AnalysisSpec.from_dict(toml.loads(conf))
        cfg = spec.resolve(experiments[0], config_collection)
        assert cfg.experiment.start_date == dt.datetime(2020, 12, 31, tzinfo=pytz.utc)
        assert cfg.experiment.end_date == dt.datetime(2021, 2, 1, tzinfo=pytz.utc)

    def test_enrollment_end_date(self, experiments, config_collection):
        conf = dedent(
            """
            [experiment]
            """
        )
        spec = AnalysisSpec.from_dict(toml.loads(conf))
        cfg = spec.resolve(experiments[7], config_collection)
        assert cfg.experiment.enrollment_period == 3

    def test_enrollment_end_date_overwrite(self, experiments, config_collection):
        conf = dedent(
            """
            [experiment]
            enrollment_period = 8
            """
        )
        spec = AnalysisSpec.from_dict(toml.loads(conf))
        cfg = spec.resolve(experiments[7], config_collection)
        assert cfg.experiment.enrollment_period == 8

    def test_private_experiment_no_dataset(self, experiments):
        conf = dedent(
            """
            [experiment]
            is_private = true
            """
        )
        with pytest.raises(ClassValidationError):
            AnalysisSpec.from_dict(toml.loads(conf))

    def test_private_experiment(self, experiments, config_collection):
        conf = dedent(
            """
            [experiment]
            is_private = true
            dataset_id = "test"
            """
        )
        spec = AnalysisSpec.from_dict(toml.loads(conf))
        cfg = spec.resolve(experiments[7], config_collection)
        assert cfg.experiment.dataset_id == "test"

    def test_sample_size_defined_experiment(self, experiments, config_collection):
        conf = dedent(
            """
            [experiment]
            sample_size = 8
            """
        )
        spec = AnalysisSpec.from_dict(toml.loads(conf))
        cfg = spec.resolve(experiments[7], config_collection)
        assert cfg.experiment.sample_size == 8

    def test_sample_size_none_experiment(self, experiments, config_collection):
        conf = dedent(
            """
            [experiment]
            enrollment_period = 7
            """
        )
        spec = AnalysisSpec.from_dict(toml.loads(conf))
        cfg = spec.resolve(experiments[7], config_collection)
        assert cfg.experiment.sample_size is None

    def test_analysis_unit_default(self, experiments, config_collection):
        conf = dedent(
            """
            [experiment]
            enrollment_period = 7
            """
        )
        spec = AnalysisSpec.from_dict(toml.loads(conf))
        cfg = spec.resolve(experiments[1], config_collection)
        assert cfg.experiment.randomization_unit is None
        assert cfg.experiment.analysis_unit == AnalysisUnit.CLIENT

    @pytest.mark.parametrize(
        "randomization_unit", [ru for ru in RandomizationUnit.__members__.values()]
    )
    def test_analysis_unit_configured(self, config_collection, randomization_unit):
        conf = dedent(
            """
            [experiment]
            enrollment_period = 7
            """
        )
        exp = Experiment(
            experimenter_slug="test_slug",
            type="pref",
            status="Complete",
            start_date=dt.datetime(2019, 12, 1, tzinfo=pytz.utc),
            end_date=dt.datetime(2020, 3, 1, tzinfo=pytz.utc),
            proposed_enrollment=7,
            branches=[Branch(slug="a", ratio=1), Branch(slug="b", ratio=1)],
            normandy_slug="normandy-test-slug",
            reference_branch="b",
            is_high_population=False,
            app_name="firefox_desktop",
            bucket_config=BucketConfig(
                randomization_unit=randomization_unit,
                namespace="testing",
                start=0,
                count=10,
                total=100,
            ),
        )
        spec = AnalysisSpec.from_dict(toml.loads(conf))
        cfg = spec.resolve(exp, config_collection)
        expected_randomization_unit = RandomizationUnit(randomization_unit)
        expected_analysis_unit = (
            AnalysisUnit.PROFILE_GROUP
            if expected_randomization_unit == RandomizationUnit.GROUP_ID
            else AnalysisUnit.CLIENT
        )
        assert cfg.experiment.randomization_unit == expected_randomization_unit
        assert cfg.experiment.analysis_unit == expected_analysis_unit

    @pytest.mark.parametrize("randomization_unit", [None, "invalid_id"])
    def test_analysis_unit_invalid(self, config_collection, randomization_unit):
        conf = dedent(
            """
            [experiment]
            enrollment_period = 7
            """
        )
        exp = Experiment(
            experimenter_slug="test_slug",
            type="pref",
            status="Complete",
            start_date=dt.datetime(2019, 12, 1, tzinfo=pytz.utc),
            end_date=dt.datetime(2020, 3, 1, tzinfo=pytz.utc),
            proposed_enrollment=7,
            branches=[Branch(slug="a", ratio=1), Branch(slug="b", ratio=1)],
            normandy_slug="normandy-test-slug",
            reference_branch="b",
            is_high_population=False,
            app_name="firefox_desktop",
            bucket_config=BucketConfig(
                randomization_unit=randomization_unit,
                namespace="testing",
                start=0,
                count=10,
                total=100,
            ),
        )
        spec = AnalysisSpec.from_dict(toml.loads(conf))
        cfg = spec.resolve(exp, config_collection)
        with pytest.raises(ValueError):
            cfg.experiment.randomization_unit
        with pytest.raises(ValueError):
            cfg.experiment.analysis_unit


class TestDefaultConfiguration:
    def test_descriptions_defined(self, experiments, config_collection):
        default_spec = AnalysisSpec.from_dict(toml.load(DEFAULT_METRICS_CONFIG))
        cfg = default_spec.resolve(experiments[0], config_collection)
        ever_ran = False

        for summaries in cfg.metrics.values():
            for _ in summaries:
                ever_ran = True
        assert ever_ran



from pathlib import Path

import pytest

ROOT = Path(__file__).parent
TEST_DATA = ROOT / "sql"


def test_generate_query_single_metric(config_collection):
    assert (
        config_collection.get_metrics_sql(metrics=["active_hours"], platform="firefox_desktop")
        == (TEST_DATA / "test_generate_query_single_metric.expected.sql").read_text()
    )


def test_generate_query_multiple_metrics(config_collection):
    assert (
        config_collection.get_metrics_sql(
            metrics=["active_hours", "days_of_use"], platform="firefox_desktop"
        )
        == (TEST_DATA / "test_generate_query_multiple_metrics.expected.sql").read_text()
    )


def test_generate_query_with_parameters(config_collection):
    assert (
        config_collection.get_metrics_sql(
            metrics=["active_hours", "days_of_use"],
            platform="firefox_desktop",
            group_by=["build_id", "sample_id"],
            where="submission_date = '2023-01-01' AND normalized_channel = 'release'",
        )
        == (TEST_DATA / "test_generate_query_with_parameters.expected.sql").read_text()
    )

    assert (
        config_collection.get_metrics_sql(
            metrics=["active_hours", "days_of_use"],
            platform="firefox_desktop",
            group_by={"build_id": "build_id", "sample_id": "sample_id"},
            where="submission_date = '2023-01-01' AND normalized_channel = 'release'",
        )
        == (TEST_DATA / "test_generate_query_with_parameters.expected.sql").read_text()
    )


def test_generate_query_with_multiple_metrics_different_data_sources(config_collection):
    assert (
        config_collection.get_metrics_sql(
            metrics=["active_hours", "days_of_use", "unenroll", "view_about_logins"],
            platform="firefox_desktop",
            group_by=["build_id", "sample_id"],
            where="submission_date = '2023-01-01' AND normalized_channel = 'release'",
        )
        == (
            TEST_DATA
            / "test_generate_query_with_multiple_metrics_different_data_sources.expected.sql"
        ).read_text()
    )


def test_generate_query_without_client_id_submission_date(config_collection):
    assert (
        config_collection.get_metrics_sql(
            metrics=["active_hours"],
            platform="firefox_desktop",
            group_by=["build_id"],
            where="submission_date = '2023-01-01' AND normalized_channel = 'release'",
            group_by_client_id=False,
            group_by_submission_date=False,
        )
        == (TEST_DATA / "test_generate_query_without_client_id.expected.sql").read_text()
    )


def test_generate_query_with_null_client_id(config_collection):
    assert (
        config_collection.get_metrics_sql(
            metrics=["cohort_clients_in_cohort"],
            platform="firefox_desktop",
        )
        == (TEST_DATA / "test_generate_query_with_null_client_id.expected.sql").read_text()
    )


def test_no_metric_definition_found(config_collection):
    with pytest.raises(ValueError):
        config_collection.get_metrics_sql(metrics=["doesnt-exist"], platform="firefox_desktop")


def test_data_source(config_collection):
    assert (
        config_collection.get_data_source_sql(
            data_source="main",
            platform="firefox_desktop",
            where="submission_date = '2023-01-01'",
        )
        == (TEST_DATA / "test_generate_data_source.expected.sql").read_text()
    )


def test_data_source_not_found(config_collection):
    with pytest.raises(ValueError):
        config_collection.get_data_source_sql(
            data_source="non-existing", platform="firefox_desktop"
        )


def test_data_source_with_join(config_collection):
    assert (
        config_collection.get_data_source_sql(
            data_source="joined_baseline",
            platform="firefox_desktop",
            where="submission_date = '2023-01-01'",
        )
        == (TEST_DATA / "test_generate_data_source_with_join.expected.sql").read_text()
    )


def test_data_source_with_multiple_join(config_collection):
    assert (
        config_collection.get_data_source_sql(
            data_source="multiple_joined_baseline",
            platform="firefox_desktop",
            where="submission_date = '2023-01-01'",
        )
        == (TEST_DATA / "test_generate_data_source_with_multi_join.expected.sql").read_text()
    )


def test_metric_with_joined_data_source(config_collection):
    assert (
        config_collection.get_metrics_sql(
            metrics=["joined_metric"],
            platform="firefox_desktop",
            where="submission_date = '2023-01-01' AND normalized_channel = 'release'",
            group_by_client_id=True,
            group_by_submission_date=False,
        )
        == (TEST_DATA / "test_generate_query_with_joined_data_sources.expected.sql").read_text()
    )



from typing import Any, Dict, List, Mapping, Optional

import attr

from .data_source import DataSourcesSpec
from .metric import MetricDefinition, MetricReference
from .parameter import ParameterSpec
from .util import converter


@attr.s(auto_attribs=True)
class OutcomeSpec:
    """Represents an outcome snippet."""

    friendly_name: str
    description: str
    metrics: Dict[str, MetricDefinition] = attr.Factory(dict)
    default_metrics: Optional[List[MetricReference]] = attr.ib(None)
    data_sources: DataSourcesSpec = attr.Factory(DataSourcesSpec)
    parameters: ParameterSpec = attr.Factory(ParameterSpec)

    @classmethod
    def from_dict(cls, d: Mapping[str, Any]) -> "OutcomeSpec":
        params: Dict[str, Any] = {}
        params["friendly_name"] = d["friendly_name"]
        params["description"] = d["description"]
        params["data_sources"] = converter.structure(d.get("data_sources", {}), DataSourcesSpec)
        params["metrics"] = {
            k: converter.structure(
                {"name": k, **dict((kk.lower(), vv) for kk, vv in v.items())},
                MetricDefinition,
            )
            for k, v in d.get("metrics", {}).items()
        }
        params["default_metrics"] = [
            converter.structure(m, MetricReference) for m in d.get("default_metrics", [])
        ]

        params["parameters"] = ParameterSpec.from_dict(d.get("parameters", dict()))

        # check that default metrics are actually defined in outcome
        for default_metric in params["default_metrics"]:
            if default_metric.name not in params["metrics"].keys():
                raise ValueError(f"Default metric {default_metric} is not defined in outcome.")

        return cls(**params)



from textwrap import dedent

import pytest
import toml

from metric_config_parser.definition import DefinitionSpec
from metric_config_parser.monitoring import MonitoringConfiguration, MonitoringSpec


class TestMonitoringSpec:
    def test_trivial_configuration(self, config_collection):
        spec = MonitoringSpec.from_dict({})
        assert isinstance(spec, MonitoringSpec)
        cfg = spec.resolve(experiment=None, configs=config_collection)
        assert isinstance(cfg, MonitoringConfiguration)
        assert cfg.metrics == []

    def test_metric_definition(self, config_collection):
        config_str = dedent(
            """
            [project]
            metrics = ["test"]

            [metrics]
            [metrics.test]
            select_expression = "SELECT 1"
            data_source = "foo"

            [metrics.test.statistics]
            sum = {}

            [data_sources]
            [data_sources.foo]
            from_expression = "test"
            """
        )
        spec = MonitoringSpec.from_dict(toml.loads(config_str))
        assert spec.metrics.definitions["test"].select_expression == "SELECT 1"
        assert spec.data_sources.definitions["foo"].from_expression == "test"
        conf = spec.resolve(experiment=None, configs=config_collection)
        assert conf.metrics[0].metric.name == "test"
        assert conf.metrics[0].metric.data_source.name == "foo"

    def test_duplicate_metrics_are_okay(self, config_collection):
        config_str = dedent(
            """
            [project]
            metrics = ["test", "test"]

            [metrics]
            [metrics.test]
            select_expression = "SELECT 1"
            data_source = "foo"

            [metrics.test.statistics]
            sum = {}

            [data_sources]
            [data_sources.foo]
            from_expression = "test"
            """
        )
        spec = MonitoringSpec.from_dict(toml.loads(config_str))
        cfg = spec.resolve(experiment=None, configs=config_collection)
        assert len(cfg.metrics) == 1

    def test_data_source_definition(self, config_collection):
        config_str = dedent(
            """
            [project]
            metrics = ["test", "test2"]

            [metrics]
            [metrics.test]
            select_expression = "SELECT 1"
            data_source = "eggs"

            [metrics.test.statistics]
            sum = {}

            [metrics.test2]
            select_expression = "SELECT 1"
            data_source = "silly_knight"

            [metrics.test2.statistics]
            sum = {}

            [data_sources.eggs]
            from_expression = "england.camelot"

            [data_sources.silly_knight]
            from_expression = "france"
            """
        )
        spec = MonitoringSpec.from_dict(toml.loads(config_str))
        cfg = spec.resolve(experiment=None, configs=config_collection)
        test = [p for p in cfg.metrics if p.metric.name == "test"][0]
        test2 = [p for p in cfg.metrics if p.metric.name == "test2"][0]
        assert test.metric.data_source.name == "eggs"
        assert "camelot" in test.metric.data_source.from_expression
        assert test2.metric.data_source.name == "silly_knight"
        assert "france" in test2.metric.data_source.from_expression

    def test_merge(self, config_collection):
        """Test merging configs"""
        config_str = dedent(
            """
            [metrics]
            [metrics.test]
            select_expression = "SELECT 1"
            data_source = "foo"

            [metrics.test.statistics]
            sum = {}

            [metrics.test2]
            select_expression = "SELECT 2"
            data_source = "foo"

            [metrics.test2.statistics]
            sum = {}

            [data_sources]
            [data_sources.foo]
            from_expression = "test"

            [dimensions]
            [dimensions.foo]
            select_expression = "bar"
            data_source = "foo"
            """
        )
        spec = MonitoringSpec.from_dict(toml.loads(config_str))

        config_str = dedent(
            """
            [project]
            name = "foo"
            metrics = ["test", "test2"]

            [metrics]
            [metrics.test]
            select_expression = "SELECT 'd'"
            data_source = "foo"

            [metrics.test.statistics]
            sum = {}

            [data_sources]
            [data_sources.foo]
            from_expression = "bar"
            """
        )
        spec2 = MonitoringSpec.from_dict(toml.loads(config_str))
        spec.merge(spec2)
        cfg = spec.resolve(experiment=None, configs=config_collection)

        assert cfg.project.name == "foo"
        test = [p for p in cfg.metrics if p.metric.name == "test"][0]
        test2 = [p for p in cfg.metrics if p.metric.name == "test2"][0]
        assert test.metric.select_expression == "SELECT 'd'"
        assert test.metric.data_source.name == "foo"
        assert test.metric.data_source.from_expression == "bar"
        assert test2.metric.select_expression == "SELECT 2"

    def test_merge_statistic(self, config_collection):
        """Test merging configs"""
        config_str = dedent(
            """
            [metrics]
            [metrics.test]
            select_expression = "SELECT 1"
            data_source = "foo"
            type = "histogram"

            [data_sources]
            [data_sources.foo]
            from_expression = "test"
            """
        )
        spec = MonitoringSpec.from_definition_spec(DefinitionSpec.from_dict(toml.loads(config_str)))

        config_str = dedent(
            """
            [project]
            name = "foo"
            metrics = ["test"]

            [metrics.test.statistics]
            sum = {}
            """
        )
        spec2 = MonitoringSpec.from_dict(toml.loads(config_str))
        spec.merge(spec2)
        cfg = spec.resolve(experiment=None, configs=config_collection)

        assert cfg.project.name == "foo"
        test = [p for p in cfg.metrics if p.metric.name == "test"][0]
        assert test.metric.select_expression == "SELECT 1"
        assert test.metric.data_source.name == "foo"
        assert test.metric.type == "histogram"
        assert test.statistic.name == "sum"

    def test_merge_data_source(self, config_collection):
        """Test merging configs with data sources"""
        config_str = dedent(
            """
            [metrics]
            [metrics.test]
            select_expression = "SELECT 1"
            data_source = "foo"
            type = "histogram"

            [metrics.test.statistics]
            sum = {}

            [data_sources]
            [data_sources.foo]
            from_expression = "test"
            build_id_column = "test"
            """
        )
        spec = MonitoringSpec.from_definition_spec(DefinitionSpec.from_dict(toml.loads(config_str)))

        config_str = dedent(
            """
            [project]
            name = "foo"
            metrics = ["test"]

            [data_sources]
            [data_sources.foo]
            from_expression = "foo"
            """
        )

        spec2 = MonitoringSpec.from_dict(toml.loads(config_str))
        spec.merge(spec2)

        assert spec.data_sources.definitions["foo"].name == "foo"
        assert spec.data_sources.definitions["foo"].from_expression == "foo"
        assert spec.data_sources.definitions["foo"].build_id_column == "test"

        cfg = spec.resolve(experiment=None, configs=config_collection)

        test = [p for p in cfg.metrics if p.metric.name == "test"][0]
        assert test.metric.data_source.name == "foo"
        assert test.metric.data_source.from_expression == "foo"

    def test_unknown_metric_failure(self, config_collection):
        config_str = dedent(
            """
            [project]
            name = "foo"
            metrics = ["test", "test2"]

            [metrics]
            [metrics.test]
            select_expression = "SELECT 'd'"
            data_source = "foo"

            [metrics.test.statistics]
            sum = {}

            [data_sources]
            [data_sources.foo]
            from_expression = "test"
            """
        )

        with pytest.raises(ValueError) as e:
            spec = MonitoringSpec.from_dict(toml.loads(config_str))
            spec.resolve(experiment=None, configs=config_collection)

        assert "No definition for metric test2." in str(e)



from textwrap import dedent

import pytest
import toml

from metric_config_parser.function import FunctionsSpec


class TestFunction:
    def test_parse_valid_function_config(self):
        config_str = dedent(
            '''
            [functions]

            [functions.agg_sum]
            definition = "COALESCE(SUM({select_expr}), 0)"

            [functions.agg_any]
            definition = "COALESCE(LOGICAL_OR({select_expr}), FALSE)"

            [functions.agg_histogram_mean]
            definition = """
                SAFE_DIVIDE(
                    SUM(CAST(JSON_EXTRACT_SCALAR({select_expr}, "$.sum") AS int64)),
                    SUM((SELECT SUM(value) FROM UNNEST(mozfun.hist.extract({select_expr}).values)))
                )
            """
        '''
        )
        function_spec = FunctionsSpec.from_dict(toml.loads(config_str))

        assert function_spec.functions["agg_sum"].slug == "agg_sum"
        assert function_spec.functions["agg_sum"].definition("1") == "COALESCE(SUM(1), 0)"

        assert function_spec.functions["agg_any"].slug == "agg_any"
        assert (
            function_spec.functions["agg_any"].definition("1") == "COALESCE(LOGICAL_OR(1), FALSE)"
        )

        assert function_spec.functions["agg_histogram_mean"].slug == "agg_histogram_mean"

    def test_parse_invalid_function_config(self):
        config_str = dedent(
            """
            [functions]

            [functions.agg_sum]
        """
        )

        with pytest.raises(KeyError):
            FunctionsSpec.from_dict(toml.loads(config_str))
